/* custom fluido bootstrap */
/* restore normal link icon */
a.externalLink[href^="https"] {
    background: url('../images/internet-web-browser.png') right center no-repeat;
}
/* hide sourceforge logo on small screens */
@media (max-width: 979px) {
    body {
        padding-right: 0.5rem;
        padding-left: 0.5rem;
    }
    .nav-collapse .nav .nav-header {
        color: black;
    }
    .nav-collapse .nav > li > a, .nav-collapse .dropdown-menu a {
        color: #08c;
        font-weight: normal;
    }
    .nav-collapse .nav > li.active > a {
        color: white;
    }
}
.nav-header {
    color: black;
}

/* force logo size */
#bannerLeft img {
    height: 32px;
}
.navbar .brand {
    padding: 4px;
}
.brand img {
    height: 32px;
}
footer p {
    font-size: 0.7rem;
}
.breadcrumb {
    background-color: transparent;
    font-size: 0.7rem;
    border: 0;
    padding: 0.2rem;
    margin: 0;
}
body {
    background:#F1FAFF url(../images/inner-gd.png) repeat-x scroll 0 0;
}
hr {
    margin: 0.5rem;
}
#breadcrumbs {
    background-color: white;
    border:0;
}
.externalLink img {
    vertical-align: middle;
}
.section h2, .section h3 {
    border:0;
    background: transparent;
    color:black;
    font-size:17px;
}
.section h2 {
    padding:1px;
}
.section h3 {
    margin-left: 0.5rem;
    margin-top: 0;
    margin-bottom: 0;
    font-size:16px;
    font-weight:bold;
}
.section p, .section ul, .section pre {
    padding-left: 0.5rem;
}
.section pre {
    margin-left: 0.5rem;
}

.section table {
    margin-left:10px;
    border-collapse: collapse;
}
.small {
    font-size:9px;
}
.source {
    overflow:auto;
}
.floatLeft {
    float:left;
}
.notype {
    list-style-type:none;
    margin-left:10px;
    padding-left:0;
}
.about td{
    vertical-align: top;
}
a.download {
    display: block;
    background: #5ACE1D url(../images/download.png) 95% center no-repeat;
    color: white;
    font-weight: bold;
    padding: 8px;
    border-radius: 8px;
    margin: 5px;
}
a.donate {
    display: block;
    background: #5ACE1D url(../images/donate.png) 95% center no-repeat;
    color: white;
    font-weight: bold;
    padding: 8px;
    border-radius: 8px;
    margin: 5px;
}
a.donate-euro {
    display: block;
    background: #5ACE1D url(../images/donate-euro.png) 95% center no-repeat;
    color: white;
    font-weight: bold;
    padding: 8px;
    border-radius: 8px;
    margin: 5px;
}
.about p {
    padding: 0;
    margin-top: 0;
}